@@ -1,6 +1,6 @@
 /**********************************************************************
 ** Copyright (C) 2000 Trolltech AS.  All rights reserved.
-** Copyright (c) 2001 Phil Thompson <phil@river-bank.demon.co.uk>
+** Copyright (c) 2004 Phil Thompson <phil@riverbankcomputing.co.uk>
 **
 ** This file is part of Qt Designer.
 **
@@ -78,15 +78,35 @@ bool Uic::toBool( const QString& s )
     return s == "true" || s.toInt() != 0;
 }
 
-QString Uic::fixString( const QString &str )
+QString Uic::fixString( const QString &str, bool encode )
 {
-    QString s( str );
 #if QT_VERSION >= 0x030100
-    s.replace( "\\", "\\\\" );
-    s.replace( "\"", "\\\"" );
-    s.replace( "\r", "" );
-    s.replace( "\n", "\\n\"\n\"" );
+    QString s;
+    if ( !encode ) {
+	s = str;
+	s.replace( "\\", "\\\\" );
+	s.replace( "\"", "\\\"" );
+	s.replace( "\r", "" );
+	s.replace( "\n", "\\n\"\n\"" );
+    } else {
+	QCString utf8 = str.utf8();
+	const int l = utf8.length();
+
+	for ( int i = 0; i < l; ++i ) {
+	    uchar ch = (uchar)utf8[i];
+
+	    if (ch != 0x0d) { // skip \r
+		s += "\\x";
+
+		if (ch <= 0x0f)
+		    s += "0";
+
+		s += QString::number( ch, 16 );
+	    }
+	}
+    }
 #else
+    QString s( str );
     s.replace( QRegExp( "\\\\" ), "\\\\" );
     s.replace( QRegExp( "\"" ), "\\\"" );
     s.replace( QRegExp( "\r?\n" ), "\\n\"\n\"" );
@@ -100,29 +120,28 @@ QString Uic::trcall( const QString& sourceText, const QString& comment )
         return "QString.null";
 
     QString t = trmacro;
+    bool encode = FALSE;
     if ( t.isNull() ) {
-	bool utf8 = FALSE;
-
         t = "self.__tr";
         for ( int i = 0; i < (int) sourceText.length(); i++ ) {
             if ( sourceText[i].unicode() >= 0x80 ) {
                 t = "self.__trUtf8";
-		utf8 = TRUE;
+		encode = TRUE;
                 break;
             }
         }
 
-	if (utf8)
+	if (encode)
 	    pyNeedTrUtf8 = TRUE;
 	else
 	    pyNeedTr = TRUE;
     }
 
     if ( comment.isEmpty() ) {
-        return t + "(" + fixString( sourceText ) + ")";
+        return t + "(" + fixString( sourceText, encode ) + ")";
     } else {
-        return t + "(" + fixString( sourceText ) + "," +
-               fixString( comment ) + ")";
+        return t + "(" + fixString( sourceText, encode ) + "," +
+               fixString( comment, encode ) + ")";
     }
 }
 
@@ -141,9 +160,9 @@ QString Uic::mkStdSet( const QString& prop )
  */
 Uic::Uic( const QString &fn, const char *outputFn, QTextStream &outStream,
 	  QDomDocument doc, bool subcl, const QString &trm,
-	  const QString& subClass, bool omitForwardDecls, QString &uicClass )
+	  const QString& subClass, QString &uicClass )
     : out( outStream ), trout (&languageChangeBody ),
-      outputFileName( outputFn ), trmacro( trm ), nofwd( omitForwardDecls ),
+      outputFileName( outputFn ), trmacro( trm ),
       pyNeedTr(FALSE), pyNeedTrUtf8(FALSE)
 {
     fileName = fn;
@@ -163,6 +182,9 @@ Uic::Uic( const QString &fn, const char *outputFn, QTextStream &outStream,
     pixmapLoaderFunction = getPixmapLoaderFunction( doc.firstChild().toElement() );
     nameOfClass = getFormClassName( doc.firstChild().toElement() );
 
+#if QT_VERSION >= 0x030300
+    uiFileVersion = doc.firstChild().toElement().attribute("version");
+#endif
     stdsetdef = toBool( doc.firstChild().toElement().attribute("stdsetdef") );
 
     if ( doc.firstChild().isNull() || doc.firstChild().firstChild().isNull() )
@@ -186,7 +208,7 @@ Uic::Uic( const QString &fn, const char *outputFn, QTextStream &outStream,
             defSpacing = e.attribute( "spacing", defSpacing.toString() );
             defMargin = e.attribute( "margin", defMargin.toString() );
         } else if ( e.tagName() == "layoutfunctions" ) {
-            defSpacing = e.attribute( "spacing", defSpacing.toString() );
+	    defSpacing = e.attribute( "spacing", defSpacing.toString() );
 
             bool ok;
             defSpacing.toInt( &ok );
@@ -380,6 +402,10 @@ void Uic::createActionImpl( const QDomElement &n, const QString &parent )
 	else
 	    continue;
 	bool subActionsDone = FALSE;
+#if QT_VERSION >= 0x030300
+	bool hasMenuText = FALSE;
+	QString actionText;
+#endif
 	for ( QDomElement n2 = ae.firstChild().toElement(); !n2.isNull(); n2 = n2.nextSibling().toElement() ) {
 	    if ( n2.tagName() == "property" ) {
 		bool stdset = stdsetdef;
@@ -398,6 +424,13 @@ void Uic::createActionImpl( const QDomElement &n, const QString &parent )
 		else
 		    call += "setProperty(\"" + prop + "\",QVariant(" + value + "))";
 
+#if QT_VERSION >= 0x030300
+		if (prop == "menuText")
+		    hasMenuText = TRUE;
+		else if (prop == "text")
+		    actionText = value;
+#endif
+
                 if ( n2.firstChild().toElement().tagName() == "string" ) {
                     trout << trindent << call << endl;
                 } else {
@@ -408,6 +441,11 @@ void Uic::createActionImpl( const QDomElement &n, const QString &parent )
 		subActionsDone = TRUE;
 	    }
 	}
+#if QT_VERSION >= 0x030300
+	// workaround for loading pre-3.3 files expecting bogus QAction behavior
+	if (!hasMenuText && !actionText.isEmpty() && uiFileVersion < "3.3")
+	    trout << indent << "self." << objName << ".setMenuText(" << actionText << ")" << endl;
+#endif
     }
 }
 
@@ -448,10 +486,10 @@ void Uic::createToolbarImpl( const QDomElement &n, const QString &parentClass, c
 		if ( n2.attribute( "class" ) != "Spacer" ) {
 			createObjectImpl( n2, "QToolBar", "self." + objName );
 		} else {
-		    QString child = createSpacerImpl( n, parentClass, parent, objName );
-		    out << indent << "QApplication.sendPostedEvents(" << objName
+		    QString child = createSpacerImpl( n2, parentClass, parent, objName );
+		    out << indent << "QApplication.sendPostedEvents(self." << objName
 			<< ",QEvent.ChildInserted)" << endl;
-		    out << indent << objName << ".boxLayout().addItem(" << child << ")" << endl;
+		    out << indent << "self." << objName << ".boxLayout().addItem(" << child << ")" << endl;
 		}
 	    }
 	}
@@ -464,20 +502,54 @@ void Uic::createMenuBarImpl( const QDomElement &n, const QString &parentClass, c
     out << indent << "self." << objName << " = QMenuBar(self,\"" << objName << "\")" << endl;
     createObjectImpl( n, parentClass, parent );
 
-    QDomNodeList nl = n.elementsByTagName( "item" );
-    for ( int i = 0; i < (int) nl.length(); i++ ) {
-	QDomElement ae = nl.item( i ).toElement();
-	QString itemName = ae.attribute( "name" );
-	out << indent << "self." << itemName << " = QPopupMenu(self)" << endl;
-	for ( QDomElement n2 = ae.firstChild().toElement(); !n2.isNull(); n2 = n2.nextSibling().toElement() ) {
-	    if ( n2.tagName() == "action" )
-		out << indent << "self." << n2.attribute( "name" ) << ".addTo(self." << itemName << ")" << endl;
-	    else if ( n2.tagName() == "separator" )
-		out << indent << "self." << itemName << ".insertSeparator()" << endl;
+    int i = 0;
+    QDomElement c = n.firstChild().toElement();
+    while ( !c.isNull() ) {
+	if ( c.tagName() == "item" ) {
+	    QString itemName = "self." + c.attribute( "name" );
+	    out << endl;
+	    out << indent << itemName << " = QPopupMenu(self)" << endl;
+	    createPopupMenuImpl( c, parentClass, itemName );
+	    out << indent << "self." << objName << ".insertItem(QString(\"\")," << itemName << "," << i << ")" << endl;
+	    QString findItem("self." + objName + ".findItem(%1)");
+	    findItem = findItem.arg(i);
+	    trout << indent << "if " << findItem << ":" << endl;
+	    ++indent;
+	    trout << indent << findItem << ".setText(" << trcall( c.attribute( "text" ) ) << ")" << endl;
+	    --indent;
+	} else if ( c.tagName() == "separator" ) {
+	    out << endl;
+	    out << indent << "self." << objName << ".insertSeparator(" << i << ")" << endl;
 	}
-	out << indent << "self." << objName << ".insertItem(QString(\"\"),self." << itemName << "," << i << ")" << endl;
-	trout << trindent << "self." << objName << ".findItem(" << i << ").setText(" << trcall( ae.attribute( "text" ) ) << ")" << endl;
-	out << endl;
+	c = c.nextSibling().toElement();
+	i++;
+    }
+}
+
+void Uic::createPopupMenuImpl( const QDomElement &e, const QString &parentClass, const QString &parent )
+{
+    int i = 0;
+    for ( QDomElement n = e.firstChild().toElement(); !n.isNull(); n = n.nextSibling().toElement() ) {
+	if ( n.tagName() == "action" || n.tagName() == "actiongroup" ) {
+	    QDomElement n2 = n.nextSibling().toElement();
+	    if ( n2.tagName() == "item" ) { // the action has a sub menu
+		QString itemName = "self." + n2.attribute( "name" );
+		QString itemText = n2.attribute( "text" );
+		out << indent << itemName << " = QPopupMenu(self)" << endl;
+		out << indent << parent << ".setAccel(" << trcall( n2.attribute( "accel" ) ) << ",";
+		out << parent << ".insertItem(" << "self." << n.attribute( "name" ) << ".iconSet(),";
+		out << trcall( itemText ) << "," << itemName << "))" << endl;
+		trout << indent << parent << ".changeItem(" << parent << ".idAt(" << i << "),";
+		trout << trcall( itemText ) << ")" << endl;
+		createPopupMenuImpl( n2, parentClass, itemName );
+		n = n2;
+	    } else {
+		out << indent << "self." << n.attribute( "name" ) << ".addTo(" << parent << ")" << endl;
+	    }
+	} else if ( n.tagName() == "separator" ) {
+	    out << indent << parent << ".insertSeparator()" << endl;
+	}
+	++i;
     }
 }
 
@@ -669,7 +741,7 @@ QString Uic::createListViewColumnImpl( const QDomElement &e, const QString &pare
     QString s;
     s = indent + parent + ".addColumn(" + trcall( txt, com ) + ")\n";
     if ( !pix.isEmpty() )
-	s += indent + parent + ".header().setLabel(" + parent + ".header().count() - 1," + pix + "," + trcall( txt, com ) + ")\n";
+	s += indent + parent + ".header().setLabel(" + parent + ".header().count() - 1,QIconSet(" + pix + ")," + trcall( txt, com ) + ")\n";
     if ( !clickable )
 	s += indent + parent + ".header().setClickEnabled(0," + parent + ".header().count() - 1)\n";
     if ( !resizable )
@@ -767,8 +839,8 @@ QString Uic::createLayoutImpl( const QDomElement &e, const QString& parentClass,
     layoutObjects += objName;
 
     QString margin = DomTool::readProperty( e, "margin", defMargin ).toString();
-    QString spacing = DomTool::readProperty( e, "spacing", defSpacing ).toString
-();
+    QString spacing = DomTool::readProperty( e, "spacing", defSpacing ).toString();
+    QString resizeMode = DomTool::readProperty( e, "resizeMode", QString::null ).toString();
 
     QString optcells;
     if ( isGrid )
@@ -791,6 +863,8 @@ QString Uic::createLayoutImpl( const QDomElement &e, const QString& parentClass,
 	}
 	out << "," << optcells << margin << "," << spacing << ",\"" << objName << "\")" << endl;
     }
+    if ( !resizeMode.isEmpty() )
+	out << indent << objName << ".setResizeMode(QLayout." << resizeMode << ")" << endl;
 
     if ( !isGrid ) {
 	for ( n = e.firstChild().toElement(); !n.isNull(); n = n.nextSibling().toElement() ) {
@@ -850,7 +924,7 @@ QString Uic::createSpacerImpl( const QDomElement &e, const QString& /*parentClas
     QDomElement n;
     QString objClass, objName;
     objClass = e.tagName();
-    objName = registerObject( "spacer" );
+    objName = registerObject( getObjectName( e ) );
 
     QSize size = DomTool::readProperty( e, "sizeHint", QSize( 0, 0 ) ).toSize();
     QString sizeType = DomTool::readProperty( e, "sizeType", "Expanding" ).toString();
